/* *******************************************************************
   SAF - Store and Forward

   TODO: einige magic numbers (64 und 13)

   ***************************************************************** */

// Timer to check SAF storage and forward data to external server if needed
void timerSafForward()
{
  const uint32_t interval = 10 * 1000UL;
  static uint32_t previousMillis = 0;
  if (millis() - previousMillis > interval)
  {
    previousMillis = millis();
    int result = safTriggerForward();
    if (result == 1) previousMillis = millis() - interval + 1000; // if successfull try again next second
  }
}


// Timer to do measurements and create a Store-And-Forward file
// used only for testing purposes
void timerSafStore()
{
  const uint32_t interval = 30 * 1000UL;
  static uint32_t previousMillis = 0;
  if (millis() - previousMillis > interval)
  {
    previousMillis = millis();
    safTriggerStore();
  }
}

// gather data an store data
void safTriggerStore()
{
  if (timeStatus() != timeNotSet)  // we only store information if we have a current time available
  {
    const uint16_t MESSAGE_SIZE_MAX = 64;                           // maximum bytes for Message Buffer
    char message[MESSAGE_SIZE_MAX];                                 // the temporary sending message - html body
    createPayload(message);                                         // reuse function from webclient to gather data
    //Serial.println(message);
    safStore(message);
  }
  else
    Serial.println(F("[SAF]I: Time not set - can't log to SAF"));
}

// try to forward the oldest SAF file to the external server
int safTriggerForward()
{
  char filename[fileLen];
  strcpy(filename, oldestSafFileDefault);
  uint32_t safFiles = safCountFile(filename);
  //Serial.print(F("[saf]D055 safFiles=")); Serial.println(safFiles);
  int result = 0;
  if (safFiles)
  {
    Serial.print(F("[saf]D059 oldest=")); Serial.println(filename);
    result = safForward(filename);
  }
  return result;
}

// forward a specific SAF file to the external server
// negative readed files but not transmitted
// 0 no file
// 1 positive transmitted file
// 9 client error
int safForward(char (&filename)[fileLen])
{
  Serial.println(F("[saf]readSaf()"));
  byte result = 0;
  if (!strcmp(filename, oldestSafFileDefault))    // just in case someone has called readSaf with FFFFFFFF.SAF
  {
    Serial.println(F("[saf]I: SAF empty"));
    return 0;
  }
  File dataFile = SD.open(filename);
  if (dataFile)
  {
    // read from the file until there's nothing else in it:
    size_t i = 0;
    char message[64] {""};            // MISSING: magic number
    while (dataFile.available()) {
      message[i++] = dataFile.read();
      if (i >= 64) break;             // MISSING: magic number
    }
    // close the file:
    dataFile.close();
    result = sendMessage(message);    // transmit data
    //Serial.print(F("[saf]D075 result=")); Serial.println(result);
    if (result == 1)
    {
      Serial.print(F("[saf]I: delete ")); Serial.println(filename);
      SD.remove(filename);
    }
  }
  else
  {
    // if the file didn't open, print an error:
    Serial.print(F("[saf]E: error opening ")); Serial.println(filename);
  }
  Serial.print(F("[saf]D105 result=")); Serial.println(result);
  return result;
}

// return number of SAF files on SD and the filename of oldest file
uint32_t safCountFile(char (&filename)[fileLen])
{
  File root;
  uint32_t safFileCounter = 0;
  root = SD.open("/");
  listDirectory(root, 0, safFileCounter, filename);
  if (!strcmp(filename, oldestSafFileDefault))
  {
    Serial.println(F("[saf]I: no SAF file"));
  }
  else
  {
    Serial.print(F("[saf]D122 filename=")); Serial.println(filename);
  }
  return safFileCounter;
}

// list directory and update filecounter and filename
// subfunction for countSafFile
void listDirectory(File dir, int numTabs, uint32_t &counter, char (&filename)[fileLen]) {
  while (true) {
    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (entry.isDirectory()) {
      Serial.println("/");
      listDirectory(entry, numTabs + 1, counter, filename);
    } else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
      if (strstr((entry.name() + (strlen(entry.name()) - 4)), ".SAF"))
      {
        //Serial.println(F("that's a SAF file"));
        counter++;
        if (strcmp(entry.name(), filename) <= -1) strcpy(filename, entry.name());
      }
    }
    entry.close();
  }
}


// generate a new SAF file on SD
// MISSING umbauen auf reference
void safStore(char * message)
{
  Serial.println(F("[saf]writeSaf()"));
  //const uint16_t MESSAGE_SIZE_MAX = 64;                          // maximum bytes for Message Buffer
  //char message[MESSAGE_SIZE_MAX];
  char val[30] = {'\0'};          // max filename 8.3 + 1 Nulllimiter
  uint32_t epoche = now();
  ultoa(epoche, val, 16);         // filename can only hold 8.3, hence we convert the timestamp to HEX
  strcat(val, ".SAF");
  File dataFile = SD.open(val, FILE_WRITE);         // MISSING to be checked to avoid the append if file would exist
  // if the file is available, write to it:
  if (dataFile) {
    Serial.println(F("[saf]D130 write SAF file"));
    //uint16_t messageLength = strlen(message);
    Serial.print(F("[saf]D132 strlen(message)=")); Serial.println(strlen(message));
    dataFile.print(message);
    dataFile.close();
    Serial.println(F("[saf]D113 Data was written to SAF"));
  }
  // if the file isn't open, pop up an error:
  else {
    Serial.println(F("[saf]E: error opening new SAF file"));
  }
}

/*
  // generate a new SAF file on SD
  void writeSaf_OLD()
  {
  Serial.println(F("[SAF]writeSaf()"));
  const uint16_t MESSAGE_SIZE_MAX = 64;                          // maximum bytes for Message Buffer
  char message[MESSAGE_SIZE_MAX];
  char val[30] = {'\0'};          // max filename 8.3 + 1 Nulllimiter
  uint32_t epoche = now();
  ultoa(epoche, val, 16);         // filename can only hold 8.3, hence we convert the timestamp to HEX
  strcat(val, ".SAF");
  File dataFile = SD.open(val, FILE_WRITE);
  // if the file is available, write to it:
  if (dataFile) {
    Serial.println(F("D092 write SAF file"));

    // payload to transmit MISSING: should be reused in client
    strcpy(message, "board=");
    strcat(message, TXT_BOARDID);
    strcat(message, "&sdeov=");
    ultoa(epoche, val, 10);
    strcat(message, val);
    strcat(message, "&sgem=");
    itoa(analogRead(analogPin[0]), val, 10);
    strcat(message, val);
    strcat(message, "&scem=");
    itoa(analogRead(analogPin[1]), val, 10);
    strcat(message, val);

    //uint16_t messageLength = strlen(message);
    Serial.print(F("D110 strlen(message)=")); Serial.println(strlen(message));
    dataFile.print(message);               // now() is a unsigned long, we need convert it to a string by base of 10
    dataFile.close();
    Serial.println(F("D113 Data was written to SAF"));
  }
  // if the file isn't open, pop up an error:
  else {
    Serial.println(F("E: error opening new SAF file"));
  }
  }
*/

/*
  // MISSING return by buffer, count
  void getOldestSafFile()
  {
  File root;
  strcpy (oldestSafFile, oldestSafFileDefault);  // reset file information
  root = SD.open("/");
  printDirectory(root, 0);
  if (!strcmp(oldestSafFile, oldestSafFileDefault))
  {
    Serial.println(F("[saf]I: no SAF file"));
  }
  else
  {
    Serial.print(F("[saf]oldestSafFile=")); Serial.println(oldestSafFile);
  }
  }


  // list directory and update global oldestSafFile
  void printDirectory(File dir, int numTabs) {
  while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1);
    } else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
      if (strstr((entry.name() + (strlen(entry.name()) - 4)), ".SAF"))
      {
        //Serial.println(F("that's a SAF file"));
        if (strcmp(entry.name(), oldestSafFile) <= -1) strcpy(oldestSafFile, entry.name());
      }
    }
    entry.close();
  }
  }
*/
