/* *******************************************************************
  Arduino Webclient SD NTP SAF
  - Webclient posts data to external Server
  - if transmission fails, backup data to SD Card
  - time keeping with SNTP / NTP (to optain date/time for SD card logging)

  This version:
  - webclient to post data to an external webserver
  - Store and Forward - use a FIFO buffer for client messages if the webserver doesn't respond
  - get time from (S)NTP

  description:
  - en: http://werner.rothschopf.net/microcontroller/202208_arduino_webclient_sd_saf_ntp.htm

  Hardware:
  - this sketch needs lot of SRAM. You should use an Arduino MEGA to test.

  by noiasca
  2022-08-03 published
  2022-07-24 Proof of concept for SAF is working
* *************************************************************** */

#include <SPI.h>                       // needed for Ethernet and SD Card
#include <Ethernet.h>                  // default Arduino library
#include <SD.h>                        // default Arduino library
#include <StreamLib.h>                 // download from library manager
#include <TimeLib.h>                   // download from library manager

#include "config.h"                    // a file with your configuration

#define CLIENT_DEBUG_USART 0                              // activate DEBUG messages for the webclient

// client
EthernetClient client;

// NTP
EthernetUDP Udp;                                 // create a UDP object for NTP
constexpr int NTP_PACKET_SIZE = 48;              // is 48
byte packetBuffer[NTP_PACKET_SIZE];              // buffer to hold incoming and outgoing packets

// SAF
constexpr byte fileLen {13};                     // Filename 8+1+3+1 = 13 bytes.
constexpr char oldestSafFileDefault[fileLen] {"FFFFFFFF.SAF"};  // Filename 8+1+3+1 = 13 bytes. This variable

void setup()
{
  Serial.begin(115200);
  Serial.println(F("\nSimple Webclient with Backup to SD, NTP, SAF"));
  for (auto &pin : ledPin) {
    pinMode(pin, OUTPUT);                         // define the digital pins to write to
  }
  pinMode(ssETH, OUTPUT); 
  digitalWrite(ssETH, HIGH);                      // unselect Ethernet shield manually
  pinMode(ssSD, OUTPUT); 
  digitalWrite(ssSD, HIGH);                       // unselect SD Card manually
  ethernetBegin();                                // start the Ethernet
  sdSetup();                                      // start the SD card

  Udp.begin(localPort);                           // start UDP for NTP/Time
}

void loop()
{
  timerNTP(false);           // checks if it is time to sync with NTP
  timerClient();             // checks if it is time to send data via client
  //timerSafStore();         // * checks if it is time to create a SAF file
  timerSafForward();         // check SAF if something ist to be forwarded
  serialUpdate();            // to control via serial interface
}
